/* =============================================================================
 UPC - EETAC - CSD - http://digsys.upc.edu 
  
 * Integrating MPLAB X + XC8 + MCC + Proteus VSM 
 
 * Explorer 8 board user guide: 
        https://ww1.microchip.com/downloads/en/DeviceDoc/40001812B.pdf   

 * Materials adapted from interrupt example code at: 
        https://digsys.upc.edu/csd/P10/P10.html#exp8 
        
 * Design phase #1: Prototype adaptation from LAB10 laboratory plan X 
        https://digsys.upc.edu/csd/P10/lab10/lab10.html#PlanX

An example on how to implement a sequential system like an 1-digit BCD counter 
using "current_state" state-type variable. 
Thus, this example can be used to adapt other FSM containing a small number 
of states.

 * Target microcontroller: PIC18F46k22 
        https://www.microchip.com/en-us/product/PIC18F46K22 

 * Print in colour this file, and analyse it and use it as 'theory' class notes. 
  
============================================================================= */

/* 	============================================================================
    Board button S1 is INT0 and it is the external counter's CLK.  
    S2 button is CE_L control input.
    LEDs RB3 RB2


  * Explorer 8 LCD display is installed by default. It will be used in next 
    design phase #2. 
   	========================================================================= */

/* 	============================================================================
 IMPORTANT NOTE: This code is used with MCC (MPLAB code configurator) 
 * Do not update or force update main.c and main.h files because 
 * Be aware of what to update for other c and h files generated by MCC
   	========================================================================= */

// MCC generated application headers 
#include "mcc_generated_files/mcc.h"
#include "lcd.h"
// Application header including global variables and function prototypes
#include "main.h" //

/* 	============================================================================
	Main function 
  	========================================================================= */
void main(void){
    init_system ();
    while(1) {                      /* loop forever the FSM */
		if (state_logic()) break;
		if (output_logic()) break;
    }
    while (1){
        __nop();//The system has crashed in an illegal state and needs attention
                //This second infinite loop is never reached in normal operation
    }           // The only way to scape is clicking reset (CD)
}	


/* 	============================================================================
	Function definitions
  	========================================================================= */

//*****************************************************************************
//State variable logic routine  (The CC1)
//*****************************************************************************
char state_logic(void){
	char error = 0;	

    read_inputs(); //sample input values so that the CC1 can decide transitions
    
	if (Var_CLK_flag == 1){

    	switch (current_state) {
            case Num_0:				
                if (Var_CE == 1){
                    current_state = Num_1;                         
                }
                else{
                    current_state = Num_0;  
                }                    
            break;

            case Num_1:				
                if (Var_CE == 1) current_state = Num_2;                         
                else current_state = Num_1;                    
            break;
            
            case Num_2:		//else is not required. Check if you see 
                            // differences in the disassembly listing 
                if (Var_CE == 1) 
                    current_state = Num_3;                         
            break;

            case Num_3:				
                if (Var_CE == 1) 
                    current_state = Num_4;    
            break;
            case Num_4:				
                if (Var_CE == 1) 
                    current_state = Num_5;    
            break;
            case Num_5:				
                if (Var_CE == 1) 
                    current_state = Num_6;    
            break;
            case Num_6:				
                if (Var_CE == 1) 
                    current_state = Num_7;    
            break;
            case Num_7:				
                if (Var_CE == 1) 
                    current_state = Num_8;    
            break;
            case Num_8:				
                if (Var_CE == 1) 
                    current_state = Num_9;    
            break;
            case Num_9:				
                if (Var_CE == 1) 
                    current_state = Num_0;    
            break;

            default:
            // If no states have been selected, there is an error
            // and the systems must be stopped                     
                error = 1;	
                break;
        }
		Var_CLK_flag = 0;	// Clear the Var_CLK_flag detection flag
	}
	return (error);
}

//*****************************************************************************
//Output logic routine (The combinational circuit CC2)
//*****************************************************************************
char output_logic(void){
unsigned char error = 0;	 
	switch (current_state) {	
		case Num_0:
            Var_BCD_code = 0;
			break;
		case Num_1:
            Var_BCD_code = 1;
			break;
		case Num_2:
            Var_BCD_code = 2;
			break;
		case Num_3:
            Var_BCD_code = 3;
			break;
		case Num_4:
            Var_BCD_code = 4;
			break;
		case Num_5:
            Var_BCD_code = 5;
			break;
		case Num_6:
            Var_BCD_code = 6;
			break;
		case Num_7:
            Var_BCD_code = 7;
			break;
		case Num_8:
            Var_BCD_code = 8;
			break;
		case Num_9:
            Var_BCD_code = 9;
			break;
		default:
			error = 1;	// If no states have been selected, there is an error
						// and the systems must be stopped 
			break;
	}
// Notice that '&' is a bitwise AND operation, however '&&' is a logic AND: 
	if (current_state == Num_9 && Var_CE == 1)
		Var_TC10 = 1;
	else
		Var_TC10 = 0;

    write_outputs(); // The interface with the hardware to convert variables 
					 // into voltages at a given microcontroller pin. 
    
	return (error);

}

/* =============================================================================
 Scan inputs and process the bits so that we can set our convenient variables 
 many of the variables are "char" because we can easily monitor them using
 the watch window for debugging purposes.
  =========================================================================== */
void read_inputs (void){     
// Like in P9, read the port, mask the bit of interest and set the variable 
    uint8_t var_buf; 
    var_buf = (PORTA & 0b00100000)>>5; //Active_low value read
    Var_CE = (!var_buf)& 0b00000001; // bitwise not
    // S2 connected at PORTA bit 5 is CE_L
}

/* =============================================================================
	Driving the output pins. Let's drive the output ports, converting 
	microcontroller variables into electrical signals. 
 * In Explorer 8 we have 8 LED.
 * PORTD(3..0) will be Q(3..0) the counter's BCD output code   
 * PORTB3 will be the counter's terminal count TC10
   ========================================================================== */
void write_outputs(void){
	uint8_t Var_buf; 
    //To write a given output pin:
    // Read and mask to preserve all the bits not to be written
    Var_buf = PORTB & 0b11110111; 
    // Shift the bit of interest to the pin position, compose the port bits (OR)
    // and write the port
    PORTB = Var_buf | (Var_TC10 << 3);  // TC10 is PORTB bit 3

    //The same procedure to write the Q(3..0) to the assigned PORTD pins (3..0)
    Var_buf = PORTD & 0b11110000; 
    PORTD = Var_buf | Var_BCD_code;
}


/* =============================================================================
	Init system: Initialise the microcontroller system 
   ========================================================================== */
void init_system (void){
    
    SYSTEM_Initialize();  //Configuration from MCC
    LCD_Initialize();     // Default LCD on Explorer 8
    
    // Enable global interrupts to use INT0
    INTERRUPT_GlobalInterruptEnable(); 
    
    // Initial message to check that hardware and software is ok
    LCD_GoTo(0,0);
    LCD_WriteString("    CSD         ");
    LCD_GoTo(1,0);
    LCD_WriteString("  Explorer 8    ");
    
}